@@ -5,13 +5,12 @@ import logging
 import os
 import typing
 import urllib.parse
-import uuid
 
 import flask
 import werkzeug.exceptions as http_error
 from rop import read_only_properties
 
-from . import disposition, from_config, utils
+from . import disposition, from_config, tokens, utils
 
 LOGGER = logging.getLogger(__name__)
 
@@ -39,15 +38,13 @@ def load_template(filename: str) -> str:
 
 def redir_dest_to_path(destination: str):
     """ Convert a redirection destination to a path fragment """
-    if destination.startswith('/'):
-        return destination[1:]
-    return destination
+    assert destination.startswith('/'), "Redirection destinations must begin with '/'"
+    return destination[1:]
 
 
 def redir_path_to_dest(path: str):
     """ Convert a path fragment to a redirection destination """
-    if path.startswith('/'):
-        return path
+    assert not path.startswith('/'), "Path fragments cannot start with '/'"
     return '/' + path
 
 
@@ -72,8 +69,9 @@ class AuthlFlask:
                  stylesheet: str = None,
                  on_verified: typing.Callable = None,
                  make_permanent: bool = True,
-                 token_storage=flask.session,
-                 token_private_namespace='_authl',
+                 state_storage: dict = flask.session,
+                 token_storage: tokens.TokenStore = None,
+                 session_namespace='_authl',
                  ):
         """ Setup Authl to work with a Flask application.
 
@@ -107,10 +105,15 @@ class AuthlFlask:
             after setting the session value)
         :param bool make_permanent: Whether a session should persist past the
             browser window closing
-        :param token_storage: The mechanism to use for token storage, for login
-            methods that need it. Defaults to using the Flask session.
-        :param token_private_namespace: The session namespace for Authl to store
-            internal data in. Only change this if you have a really good reason.
+        :param tokens.TokenStore token_storage: Storage for token data for
+            methods which use it. Defaults to tokens.Serializer using the Flask
+            app.secret_key; this is suitable for load-balancing scenarios as
+            long as all service nodes use the same secret_key.
+        :param state_storage: The mechanism to use for transactional state
+            storage for login methods that need it. Defaults to using the Flask
+            user session.
+        :param session_namespace: A namespace for Authl to keep a small amount of
+            user session data in. Should never need to be changed.
 
         The login_render_func takes the following arguments; note that more may
         be added so it should also take a **kwargs for future compatibility:
@@ -155,11 +158,12 @@ class AuthlFlask:
         """
         # pylint:disable=too-many-arguments,too-many-locals,too-many-statements
 
-        self.instance = from_config(config,
-                                    app.secret_key or uuid.uuid4().bytes,
-                                    token_storage)
+        self.instance = from_config(
+            config,
+            state_storage,
+            token_storage or tokens.Serializer(app.secret_key))
 
-        self._session = token_storage
+        self._session = state_storage
         self.login_name = login_name
         self.callback_name = callback_name
         self.tester_name = tester_name
@@ -168,16 +172,18 @@ class AuthlFlask:
         self._notify_render_func = notify_render_func
         self._session_auth_name = session_auth_name
         self.force_ssl = force_ssl
-        if stylesheet is not None:
-            setattr(self, 'stylesheet', str(stylesheet))
+        self._stylesheet = stylesheet
         self._on_verified = on_verified
         self.make_permanent = make_permanent
-        self._prefill_key = token_private_namespace + '.prefill'
+        self._prefill_key = session_namespace + '.prefill'
 
         for sfx in ['', '/', '/<path:redir>']:
             app.add_url_rule(login_path + sfx, login_name,
                              self._login_endpoint, methods=('GET', 'POST'))
-        app.add_url_rule(callback_path + '/<hid>', callback_name, self._callback_endpoint)
+        app.add_url_rule(callback_path + '/<hid>',
+                         callback_name,
+                         self._callback_endpoint,
+                         methods=('GET', 'POST'))
 
         if tester_path:
             def find_service():
@@ -319,11 +325,13 @@ class AuthlFlask:
     @property
     def stylesheet(self) -> str:
         """ Gets the stylesheet for the Flask templates """
+        if self._stylesheet:
+            return utils.resolve_value(self._stylesheet)
         return flask.url_for(self.login_name, asset='css')
 
 
 def client_id():
-    """ A shim to generate a client ID for IndieAuth/IndieLogin """
+    """ A shim to generate a client ID for IndieAuth """
     from flask import request
     parsed = urllib.parse.urlparse(request.base_url)
     baseurl = '{}://{}'.format(parsed.scheme, parsed.hostname)
