@@ -1,14 +1,14 @@
 """ Handler for emailing a magic link """
 
 import email
-import html
 import logging
 import math
+import time
 import urllib.parse
 
 import validate_email
 
-from .. import disposition, utils
+from .. import disposition, tokens, utils
 from . import Handler
 
 LOGGER = logging.getLogger(__name__)
@@ -50,12 +50,12 @@ class EmailAddress(Handler):
 
     @property
     def logo_html(self):
-        return [(utils.read_icon('email_addr.svg'), 'email')]
+        return [(utils.read_icon('email_addr.svg'), 'Email')]
 
     def __init__(self,
                  sendmail,
                  notify_cdata,
-                 token_store,
+                 token_store: tokens.TokenStore,
                  expires_time=None,
                  email_template_text=DEFAULT_TEMPLATE_TEXT,
                  please_wait_error=DEFAULT_WAIT_ERROR,
@@ -91,11 +91,11 @@ class EmailAddress(Handler):
 
     def handles_url(self, url):
         """ Validating email by regex: not even once """
-        try:
-            if urllib.parse.urlparse(url).scheme == 'mailto':
-                return url
-        except (ValueError, AttributeError):
-            pass
+        parsed = urllib.parse.urlparse(url)
+        if parsed.scheme == 'mailto' and validate_email.validate_email(parsed.path):
+            return url
+        if parsed.scheme:
+            return None
 
         if validate_email.validate_email(url):
             return 'mailto:' + url
@@ -104,9 +104,12 @@ class EmailAddress(Handler):
 
     def initiate_auth(self, id_url, callback_uri, redir):
         # Extract the destination email from the identity URL
-        dest_addr = urllib.parse.urlparse(id_url).path.lower()
+        parsed = urllib.parse.urlparse(id_url)
+        if parsed.scheme != 'mailto' or not validate_email.validate_email(parsed.path):
+            return disposition.Error("Malformed email URL", redir)
+        dest_addr = parsed.path.lower()
 
-        token = self._token_store.dumps((dest_addr, redir))
+        token = self._token_store.put((dest_addr, redir, time.time()))
 
         link_url = (callback_uri + ('&' if '?' in callback_uri else '?') +
                     urllib.parse.urlencode({'t': token}))
@@ -130,12 +133,12 @@ class EmailAddress(Handler):
             return disposition.Error('Missing token', None)
 
         try:
-            email_addr, redir = utils.unpack_token(self._token_store, token, self._lifetime)
+            email_addr, redir, when = self._token_store.pop(token)
         except disposition.Disposition as disp:
             return disp
 
-        if not email_addr or not validate_email.validate_email(email_addr):
-            return disposition.Error('Invalid email address ' + html.escape(str(email_addr)), redir)
+        if time.time() > when + self._lifetime:
+            return disposition.Error("Login timed out", redir)
 
         return disposition.Verified('mailto:' + email_addr, redir)
 
@@ -175,7 +178,7 @@ def simple_sendmail(connector, sender_address, subject):
 
     """
 
-    def sendmail(message):
+    def sendmail(message: email.message.EmailMessage):
         message['From'] = sender_address
         message['Subject'] = subject
 
@@ -185,7 +188,7 @@ def simple_sendmail(connector, sender_address, subject):
     return sendmail
 
 
-def from_config(config, token_store):
+def from_config(config, token_store: tokens.TokenStore):
     """ Generate an EmailAddress handler from the provided configuration dictionary.
 
     Possible configuration values (all optional unless specified):
