@@ -1,19 +1,25 @@
 """ IndieAuth login handler. """
 
 import logging
+import time
 import typing
 import urllib.parse
 
+import expiringdict
 import requests
 from bs4 import BeautifulSoup
 
-from .. import disposition, utils
+from .. import disposition, tokens, utils
 from . import Handler
 
 LOGGER = logging.getLogger(__name__)
 
+# We do this instead of functools.lru_cache so that IndieAuth.handles_page
+# and find_endpoint can both benefit from the same endpoint cache
+_ENDPOINT_CACHE = expiringdict.ExpiringDict(max_len=128, max_age_seconds=1800)
 
-def find_endpoint(id_url: str = None,
+
+def find_endpoint(id_url: str,
                   links: typing.Dict = None,
                   content: BeautifulSoup = None) -> typing.Optional[str]:
     """ Given an identity URL, discover its IndieAuth endpoint
@@ -22,30 +28,39 @@ def find_endpoint(id_url: str = None,
     :param links: a request.links object from a requests operation
     :param BeautifulSoup content: a BeautifulSoup parse tree of an HTML document
     """
-    def _link_endpoint(links):
-        if not links:
-            return None
-
-        LOGGER.debug("checking indieauth by link header")
-        if 'authorization_endpoint' in links:
+    def _derive_endpoint(links, content):
+        LOGGER.debug('links for %s: %s', id_url, links)
+        if links and 'authorization_endpoint' in links:
+            LOGGER.debug("Found link header")
             return links['authorization_endpoint']['url']
+
+        if content:
+            link = content.find('link', rel='authorization_endpoint')
+            if link:
+                LOGGER.debug("Found link tag")
+                return urllib.parse.urljoin(id_url, link.get('href'))
+
         return None
 
-    def _content_endpoint(content):
-        if not content:
-            return None
+    # Get the cached endpoint value, but don't immediately use it if we have
+    # links and/or content, as it might have changed
+    cached = _ENDPOINT_CACHE.get(id_url)
+    LOGGER.debug("Cached endpoint for %s: %s", id_url, cached)
 
-        LOGGER.debug("checking indieauth by link tag")
-        link = content.find('link', rel='authorization_endpoint')
-        if link:
-            return link.get('href')
-        return None
+    found = (links or content) and _derive_endpoint(links, content)
+    if id_url and not found and not cached:
+        # We didn't find a new endpoint, and we didn't have a cached one
+        LOGGER.debug("Retrieving %s", id_url)
+        request = utils.request_url(id_url)
+        found = _derive_endpoint(request.links,
+                                 BeautifulSoup(request.text, 'html.parser'))
 
-    request = utils.request_url(id_url) if id_url and not links or not content else None
-    if request:
-        links = request.links
-        content = BeautifulSoup(request.text, 'html.parser')
-    return _link_endpoint(links) or _content_endpoint(content)
+    if found and id_url:
+        # we found a new value so update the cache
+        LOGGER.debug("Caching %s -> %s", id_url, found)
+        _ENDPOINT_CACHE[id_url] = found
+
+    return found or cached
 
 
 def verify_id(request_id: str, response_id: str) -> typing.Optional[str]:
@@ -56,6 +71,10 @@ def verify_id(request_id: str, response_id: str) -> typing.Optional[str]:
     not be verified.
     """
 
+    # exact match is always okay
+    if request_id == response_id:
+        return response_id
+
     orig = urllib.parse.urlparse(request_id)
     resp = urllib.parse.urlparse(response_id)
     LOGGER.debug('orig=%s resp=%s', orig, resp)
@@ -107,9 +126,12 @@ class IndieAuth(Handler):
     """ Directly support login via IndieAuth, without requiring third-party
     IndieLogin brokerage.
 
-    IndieAuth is just barely different enough from baseline OAuth that it's
-    easier to just reimplement it directly, rather than trying to use the OAuth
-    base class.
+    SECURITY NOTE: When used with tokens.Serializer, this is subject to certain
+    classes of replay attack; for example, if the user endpoint uses irrevocable
+    signed tokens for the code grant (which is done in many endpoints, e.g.
+    SelfAuth), an attacker can replay a transaction that it intercepts. As such
+    it is very important that the timeout be configured to a reasonably short
+    time to mitigate this.
     """
 
     @property
@@ -134,7 +156,7 @@ class IndieAuth(Handler):
     def logo_html(self):
         return [(utils.read_icon('indieauth.svg'), 'IndieAuth')]
 
-    def __init__(self, client_id, token_store, timeout: int = None):
+    def __init__(self, client_id, token_store: tokens.TokenStore, timeout: int = None):
         """ Construct an IndieAuth handler
 
         :param client_id: The client_id to send to the remote IndieAuth
@@ -149,35 +171,22 @@ class IndieAuth(Handler):
         self._token_store = token_store
         self._timeout = timeout or 600
 
-        self._endpoints = utils.LRUDict()
+    def handles_url(self, url):
+        # If we already know what endpoint exists for this, go ahead and say it.
+        # Otherwise, we fall through to handles_page.
+        if url in _ENDPOINT_CACHE:
+            return url
+        return None
 
     def handles_page(self, url, headers, content, links):
-        """ If we have the appropriate link rels, register the endpoint now """
-        endpoint = find_endpoint(links=links, content=content)
-
-        if endpoint:
-            # cache this for later so we don't have to look it up again
-            LOGGER.info("%s: has IndieAuth endpoint %s", url, endpoint)
-            self._endpoints[url] = endpoint
-
-        return endpoint
-
-    def _get_endpoint(self, id_url):
-        if id_url in self._endpoints:
-            return self._endpoints[id_url]
-
-        endpoint = find_endpoint(id_url)
-        if endpoint:
-            self._endpoints[id_url] = endpoint
-
-        return endpoint
+        return find_endpoint(url, links, content) is not None
 
     def initiate_auth(self, id_url, callback_uri, redir):
-        endpoint = self._get_endpoint(id_url)
+        endpoint = find_endpoint(id_url)
         if not endpoint:
             return disposition.Error("Failed to get IndieAuth endpoint", redir)
 
-        state = self._token_store.dumps(((id_url, endpoint, callback_uri), redir))
+        state = self._token_store.put((id_url, endpoint, callback_uri, time.time(), redir))
 
         client_id = utils.resolve_value(self._client_id)
         LOGGER.debug("Using client_id %s", client_id)
@@ -192,16 +201,19 @@ class IndieAuth(Handler):
 
     def check_callback(self, url, get, data):
         # pylint:disable=too-many-return-statements
+
         state = get.get('state')
         if not state:
             return disposition.Error("No transaction provided", None)
 
         try:
-            (id_url, endpoint, callback_uri), redir = utils.unpack_token(self._token_store,
-                                                                         state, self._timeout)
+            id_url, endpoint, callback_uri, when, redir = self._token_store.pop(state)
         except disposition.Disposition as disp:
             return disp
 
+        if time.time() > when + self._timeout:
+            return disposition.Error("Transaction timed out", redir)
+
         try:
             # Verify the auth code
             request = requests.post(endpoint, data={
@@ -212,7 +224,8 @@ class IndieAuth(Handler):
 
             if request.status_code != 200:
                 LOGGER.error("Request returned code %d: %s", request.status_code, request.text)
-                return disposition.Error("Unable to verify identity", redir)
+                return disposition.Error("Authorization endpoint returned %d" % request.status_code,
+                                         redir)
 
             try:
                 response = request.json()
@@ -220,12 +233,14 @@ class IndieAuth(Handler):
                 LOGGER.error("%s: Got invalid JSON response from %s: %s (content-type: %s)",
                              id_url, endpoint,
                              request.text,
-                             request.headersversion('content-type'))
+                             request.headers.get('content-type'))
                 return disposition.Error("Got invalid response JSON", redir)
 
             response_id = verify_id(id_url, response['me'])
             if not response_id:
-                return disposition.Error("Identity URL does not match", redir)
+                return disposition.Error(
+                    "Identity URL '%s' does not match request '%s'" % (response['me'], id_url),
+                    redir)
 
             return disposition.Verified(response_id, redir, response)
         except KeyError as key:
